(*
 * Priority search queues over unbalanced trees
 * 
 *)

open EdisonPervasives

(* Ce code est bas sur le rapport technique de Ralph Hinze UU-CS 2001 09
 * mais il le reproduit plus fidlement que le module [UnbalancedPSQ]
 * en conservant notamment la simulation des vues
 *)

type 
  'a tree = 
  | E 
  | L of 'a * int * 'a tree * 'a * 'a tree
  | R of 'a * int * 'a tree * 'a * 'a tree 
and 
  'a prioritySQ = [ `Empty | `Pennant of 'a * int * 'a tree * 'a ]
    
(* les arbres de recherche sont orients feuilles *)

type 'a treeView = 
    [ `Empty
    | `Single of 'a * int 
    | `Tree of 'a prioritySQ * 'a * 'a prioritySQ ]
    
type 'a minView = [ `Empty | `Min of 'a * int * 'a prioritySQ ]


(* unit deux PSQ telles que tous les lments de la premire
 * soient infrieurs  ceux de la seconde
 *)

let link = fun a b ->
  match (a, b) with
    | (`Empty, _) -> b
    | (_, `Empty) -> a
    | (`Pennant (x, p, t, m), `Pennant (x', p', t', m')) ->
	match compare p p' with
	  | n when n <= 0 -> `Pennant (x, p, R (x', p', t, m, t'), m')
	  | _ -> `Pennant (x', p', L (x, p, t, m, t'), m')
    
(** Construction des diffrentes vues **)


let tree_view = function
  | `Empty -> `Empty
  | `Pennant (x, p, E, m) -> `Single (x, p)
  | `Pennant (x, p, L (x', p', tl, m, tr), m') ->
      `Tree (`Pennant (x', p', tl, m), m, `Pennant (x, p, tr, m'))
  | `Pennant (x, p, R (x', p', tl, m, tr), m') ->
      `Tree (`Pennant (x, p, tl, m), m, `Pennant (x', p', tr, m'))

let rec min_view_rec k = function
  | E -> `Empty
  | L (x, p, a, y, b) ->
      link (`Pennant (x, p, a, k)) (min_view_rec y b)
  | R (x, p, a, y, b) ->
      link (min_view_rec k a) (`Pennant (x, p, b, y))
      
let min_view = function
  | `Empty -> `Empty
  | `Pennant (x, p, t, m) -> `Min (x, p, min_view_rec m t)
 

(** Oprations usuelles sur les arbres pondrs **)


let unit = fun x p -> `Pennant (x, p, E, x)

let max_element = function
  | `Empty -> raise Empty
  | `Pennant (_, _, _, m) -> m

let rec insert x p = function psq  ->
  match tree_view psq with
    | `Empty -> unit x p
    | `Single (x', p') ->
	(match compare x x' with
	   | n when n < 0 -> link (unit x p) (unit x' p')
	   | 0 -> unit x p
	   | _ -> link (unit x' p') (unit x p)
	)
    | `Tree (l, m, r) ->
	match compare x m with
	  | n when n <= 0 -> link (insert x p l) r
	  | _ -> link l (insert x p r)

(* la cl m n'est qu'une cl de routage et ne peut donc pas tre efface
 * comme dans un arbre de recherche orient noeuds mais suit la procdure
 * des arbres orients feuilles (descente jusqu'en profondeur maximum)
 *)

let rec delete x = function psq ->
  match tree_view psq with
    | `Empty -> psq
    | `Single (x', p') ->
	(match compare x x' with
	   | 0 -> `Empty
	   | _ -> psq
	)
    | `Tree (l, m, r) ->
	match compare x m with
	  | n when n <= 0 -> link (delete x l) r
	  | _ -> link l (delete x r)

let rec to_list' = function psq ->
  match tree_view psq with
    | `Empty -> []
    | `Single (x, p) -> [(x, p)]
    | `Tree (l, _, r) -> to_list' l @ to_list' r

let rec to_list_rec x p = function
  | E -> [(x, p)]
  | L (x', p', l, m, r) -> to_list_rec x' p' l @ to_list_rec x p r
  | R (x', p', l, m, r) -> to_list_rec x p l @ to_list_rec x' p' r

let to_list = function
  | `Empty -> []
  | `Pennant (x, p, t, _) -> to_list_rec x p t

let rec from_list_rec result = function
  | [] -> result
  | (x, p) :: tail -> from_list_rec (insert x p result) tail
      
let from_list = function list -> from_list_rec `Empty list
  
let rec lookup x = function psq ->
  match tree_view psq with
    | `Empty -> raise Not_found
    | `Single (x', p') ->
	(match compare x x' with
	   | 0 -> p'
	   | _ -> raise Not_found
	)
    | `Tree (l, m, r) ->
	match compare x m with
	  | n when n <= 0 -> lookup x l
	  | _ -> lookup x r
	      
let zero = function () -> `Empty

let min_head = function psq ->
  match min_view psq with
    | `Empty -> raise Empty
    | `Min (x, p, _) -> (x, p)

let min_tail = function psq ->
  match min_view psq with
    | `Empty -> raise Empty
    | `Min (_, _, tail) -> tail
  
let is_empty = function
  | `Empty -> true
  | `Pennant _ -> false

let rec at_most p = function psq ->
  match min_view psq with
    | `Empty -> []
    | `Min (_, p', _) when p' > p -> []
    | _ ->
	match tree_view psq with
	  | `Empty -> []
	  | `Single (x, p) -> [(x, p)]
	  | `Tree (l, _, r) -> at_most p l @ at_most p r

let rec at_most_range p k k'= function psq ->
  match min_view psq with
    | `Empty -> []
    | `Min (_, p', _) when p' > p -> []
    | _ ->
	match tree_view psq with
	  | `Empty -> []
	  | `Single (x, p) ->
	      (match (compare x k, compare k' x) with
		| (m, n) when m >= 0 && n >= 0 -> [(x, p)]
		| _ -> []
	      )		
	  | `Tree (l, m, r) ->
	      (match compare k m with
		 | n when n <= 0 -> at_most_range p k k' l
		 | _ -> [])
	      @
	      (match compare m k' with
		 | n when n <= 0 -> at_most_range p k k' r
		 | _ -> [])
